#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a Gtk plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; A small drawing in two fix-layout widgets.
;;; package: (gtk fix-layout demo)

(define blink? #t)
(define spin? #t)

(define (make-fix-layout-widgets)
  (with-glib-lock
   (lambda ()
     (let* ((window (let ((w (gtk-window-new 'toplevel)))
		      (gtk-widget-set-opacity w 0.90)
		      (gtk-window-set-title w "fix-layout-demo")
		      (set-gtk-window-delete-event-callback!
		       w (lambda (w) (%trace ";closed "w"\n") 0))
		      (gtk-container-set-border-width w 5)
		      w))
	    (scroller1 (gtk-scrolled-view-new))
	    (scroller2 (gtk-scrolled-view-new))
	    (layout1 (let ((l (make-demo-layout 200 200)))
		       (gtk-widget-set-hexpand l #t)
		       (gtk-widget-set-vexpand l #t)
		       (set-scm-widget-minimum-size! l 40 40)
		       l))
	    (layout2 (let ((l (make-demo-layout 200 300)))
		       (gtk-widget-set-hexpand l #t)
		       (gtk-widget-set-vexpand l #t)
		       (set-scm-widget-minimum-size! l 40 60)
		       l))
	    (paned (gtk-paned-view-new 'vertical)))

       (gtk-container-add scroller1 layout1)
       (gtk-paned-pack1 paned scroller1 'resize #f)
       (gtk-container-add scroller2 layout2)
       (gtk-paned-pack2 paned scroller2 'resize #f)
       (gtk-container-add window paned)
       (gtk-widget-show-all window)
       (values layout1 layout2)))))

(define (make-fix-layout-demo)
  (with-values make-fix-layout-widgets
    (lambda (layout1 layout2)
      (let ((drawing
	     (with-glib-lock
	      (lambda ()
		(let ((drawing (make-demo-drawing layout1)))
		  (let ((cursor1 (make-cursor-ink))
			(cursor2 (make-cursor-ink)))
		    (fix-drawing-add-ink! drawing cursor1 'bottom)
		    (fix-drawing-add-ink! drawing cursor2 'bottom)
		    (set-demo-drawing-cursor-inks!
		     drawing (list (list cursor1 layout1)
				   (list cursor2 layout2))))

		  ;; Attach widgets to drawing BEFORE starting threads.
		  (set-fix-layout-drawing! layout1 drawing 75 50)
		  (set-fix-layout-drawing! layout2 drawing 75 50)

		  drawing)))))
	(if blink?
	    (let ((thread (start-blinking drawing)))
	      (%trace ";cursor blinker: "thread"\n"))
	    (%trace ";cursor blinking disabled\n"))
	(if spin?
	    (let ((thread (start-spinning drawing)))
	      (%trace ";ring spinner: "thread"\n"))
	    (%trace ";ring spinning disabled\n")))
      (with-glib-lock
       (lambda ()
	 (gtk-widget-grab-focus layout1)))
      (%trace ";created "layout1" and "layout2"\n")
      unspecific)))

(define (make-cursor-ink)
  (let ((cursor (make-rectangle-ink)))
    (set-rectangle-ink-width! cursor 2)
    (set-rectangle-ink-color! cursor "green")
    (set-rectangle-ink-fill-color! cursor (make-rgba-color 0. 1. 0. .5))
    cursor))

(define-class (<demo-layout> (constructor () (width height)))
    (<fix-layout>))

(define-method initialize-instance ((layout <demo-layout>) width height)
  (call-next-method layout width height 'white)
  (set-fix-widget-motion-handler! layout demo-motion-handler)
  (set-fix-widget-button-handler! layout 'release demo-button-release-handler)
  (set-fix-widget-key-press-handler! layout demo-key-press-handler)
  layout)

(define-method fix-widget-realize-callback ((widget <demo-layout>))
  (call-next-method widget)
  (set-fix-widget-pointer-shape! widget 'crosshair))

(define (make-demo-drawing widget)
  (let ((drawing (%make-demo-drawing)))
    (%trace ";make-demo-drawing: "drawing"\n")
    (set-fix-drawing-size! drawing 300 300)
    (let ((line1 (make-line-ink))
	  (line2 (make-line-ink))
	  (line3 (make-line-ink))
	  (poly (make-polygon-ink))
	  (arc (make-arc-ink))
	  (text (make-simple-text-ink))
	  (box (make-rectangle-ink))
	  (image (make-image-ink-from-file
		  (system-library-pathname "conses.png")))
	  (surface (make-surface-ink 40 40)))

      (set-line-ink! line1 140 150 200 150)
      (fix-drawing-add-ink! drawing line1)

      (set-line-ink! line2 150 140 150 200)
      (fix-drawing-add-ink! drawing line2)

      (set-line-ink! line3 135 150 200 85)
      (set-line-ink-width! line3 3)
      (set-line-ink-color! line3 "blue")
      (set-line-ink-dash-color! line3 "green")
      (set-line-ink-dashes! line3 '(5. 5. 10. 5.))
      (fix-drawing-add-ink! drawing line3)

      (set-text-ink-position! text 150 150)
      (set-text-ink-color! text "black")
      (set-simple-text-ink-text! text widget "Hello, World!")
      (fix-drawing-add-ink! drawing text)

      (set-rectangle-ink! box 120 120 20 20)
      (set-rectangle-ink-width! box 2)
      (set-rectangle-ink-color! box "green")
      (set-rectangle-ink-fill-color! box (make-rgba-color 0. 1. 0. .5))
      (fix-drawing-add-ink! drawing box)

      (set-image-ink! image 170 100)
      (fix-drawing-add-ink! drawing image)

      (set-arc-ink! arc 140 90 30 30)
      (set-arc-ink-width! arc 5)
      (set-arc-ink-color! arc "gold")
      (fix-drawing-add-ink! drawing arc)

      (set-polygon-ink! poly (star-vertices 15. 150. 95.))
      (set-polygon-ink-width! poly 1)
      (set-polygon-ink-color! poly "red")
      (set-polygon-ink-fill-color! poly "gold")
      (fix-drawing-add-ink! drawing poly)

      (draw-on-surface surface)
      (set-surface-ink-position! surface 175 175)
      (fix-drawing-add-ink! drawing surface)
      drawing)))

(define (star-vertices size center.x center.y)
  (let ((m (cairo-identity-matrix)))
    (cairo-matrix-scale! m size)
    (cairo-matrix-translate! m center.x center.y)
    (map (lambda (p)
	   (cairo-transform! p m)
	   (cons (round->exact (x p)) (round->exact (y p))))
	 (unit-star-vertices 5.))))

(define-integrable 2pi (flo:* 8. (flo:atan2 1. 1.)))

(define (unit-star-vertices n-points)
  (guarantee-flonum n-points 'star-vertices)
  (let* ((incr (flo:/ 2pi n-points))
	 (incr/2 (flo:/ incr 2.)))
    (let loop ((verts '())
	       (i 0.))
      (if (flo:= i n-points)
	  (reverse! (cons (cairo-point 1. 0.) verts))
	  (let ((angle (flo:* i incr)))
	    (loop `(,(cairo-point (flo:* .5 (flo:cos (flo:+ angle incr/2)))
				  (flo:* .5 (flo:sin (flo:+ angle incr/2))))
		    ,(cairo-point (flo:cos angle) (flo:sin angle))
		    ,@verts)
		  (flo:+ 1. i)))))))

(define (draw-on-surface ink)
  (let* ((surface (surface-ink-surface ink))
	 (cr (cairo-create surface))
	 (white (->color "white" 'draw-on-surface))
	 (blue (->color "blue" 'draw-on-surface)))
    (cairo-set-source-color cr white)
    (cairo-paint cr)
    (let ((extent (fix-ink-extent ink)))
      (cairo-scale cr
		   (flo:/ (->flonum (fix-rect-width extent)) 2.)
		   (flo:negate (flo:/ (->flonum (fix-rect-height extent)) 2.))))
    (cairo-translate cr 1. -1.)

    (let ((pattern #;(cairo-pattern-create-linear 0. 0. 1. 1.)
		   (cairo-pattern-create-radial .25 .25 0. .25 .25 1.)))
      (cairo-pattern-add-color-stop pattern 0. white)
      (cairo-pattern-add-color-stop pattern 1. blue)
      (cairo-arc cr 0. 0. 1. 0. (flo:* 8. (flo:atan2 1. 1.)))
      (cairo-set-source cr pattern)
      (cairo-fill cr)
      (cairo-pattern-destroy pattern)
      (cairo-destroy cr))
    (cairo-surface-flush surface)
    (drawing-damage ink)))

(define-class (<demo-drawing> (constructor %make-demo-drawing () no-init))
    (<fix-drawing>)
  ;; An alist of cursors and their widgets, for the blinking thread
  ;; and mouse motion handler.
  (cursor-inks define standard initial-value '()))

(define (demo-motion-handler layout modifiers window-x window-y)
  (%trace2 ";demo-motion-handler "layout" "modifiers" "window-x" "window-y"\n")
  (let* ((drawing (fix-layout-drawing layout))
	 (view (fix-layout-view layout))
	 (x (+ window-x (fix-rect-x view)))
	 (y (+ window-y (fix-rect-y view))))
    (%trace2 ";pointer moved to ("x","y") in "layout"\n")
    (for-each
      (lambda (ink)
	(if (not (text-ink? ink))
	    (%trace2 ";picked: "ink"\n")
	    (let ((index (text-ink-xy-to-index ink x y))
		  (text-extent (fix-ink-extent ink)))
	      (with-text-ink-grapheme-rect
	       ink index
	       (lambda (xG yG widthG heightG)
		 (for-each
		   (lambda (cursor.widgets)
		     (if (memq layout (cdr cursor.widgets))
			 (begin
			   (set-rectangle-ink!
			    (car cursor.widgets)
			    (+ xG (fix-rect-x text-extent))
			    (+ yG (fix-rect-y text-extent))
			    widthG heightG)
			   ;; Keep the cursor on while tracking the mouse.
			   (set-fix-ink-widgets!
			    (car cursor.widgets)
			    (cdr cursor.widgets)))))
		   (demo-drawing-cursor-inks drawing)))))))
      (fix-drawing-pick-list drawing layout x y)))
  #t)

(define (demo-button-release-handler layout type button modifiers window-x window-y)
  (%trace2 ";demo-button-release-handler "layout" "type" "button" "modifiers" "window-x" "window-y"\n")
  (let* ((drawing (fix-layout-drawing layout))
	 (view (fix-layout-view layout))
	 (x (+ window-x (fix-rect-x view)))
	 (y (+ window-y (fix-rect-y view))))
    (%trace ";pointer released at ("x","y")\n")
    (for-each
      (lambda (ink)
	(if (not (text-ink? ink))
	    (%trace ";picked: "ink"\n")
	    (let ((index (text-ink-xy-to-index ink x y)))
	      (%trace ";picked: "ink" (char "index")\n"))))
      (fix-drawing-pick-list drawing layout x y)))
  #t)

(define (demo-key-press-handler layout keyval char-bits)
  (%trace ";demo-key-press-handler "layout" "keyval" "char-bits"\n")
  (if (and (char? keyval) (char=? keyval #\D))
      (bkpt 'Test))
  ;;NOT handled.
  #f)

(define (start-spinning drawing)
  (let ((thread (create-spinner drawing)))
    (detach-thread thread)
    thread))

(define (create-spinner drawing)
  (create-thread
   #f
   (lambda ()
     (%trace ";spinning started\n")
     (let* ((frames 10)
	    (x 170) (y 90) (height 30) (width 30)
	    (pi (* (atan 1 1) 4))
	    (half-widths (make-vector frames))
	    (arc (find arc-ink? (fix-drawing-display-list drawing))))
       (define-integrable (half-width angle)
	 (fix:1+ (truncate->exact (/ (* width (abs (cos angle))) 2))))
       (do ((frame 0 (fix:1+ frame)))
	   ((= frame frames))
	 (vector-set! half-widths frame (half-width (* pi (/ frame frames)))))
       (let loop ((frame 0))
	 (sleep-current-thread 100)
	 (let ((half-width (vector-ref half-widths frame)))
	   (let ((x (fix:- x half-width))
		 (width (fix:* 2 half-width)))
	     (%trace2 ";spinning to "width"\n")
	     (with-glib-lock
	      (lambda () (set-arc-ink! arc x y width height)))))
	 (let ((widgets (fix-drawing-widgets drawing)))
	   (if (null? widgets)
	       (%trace ";spinning ended\n")
	       (loop (modulo (fix:1+ frame) frames)))))))
   (list 'spin drawing)))

(define (start-blinking drawing)
  (let ((thread (create-blinker drawing)))
    (detach-thread thread)
    thread))

(define (create-blinker drawing)
  (create-thread
   #f
   (lambda ()
     (%trace ";blinking started\n")
     (let loop ()
       ;; Off!
       (with-glib-lock
	(lambda ()
	  (for-each (lambda (cursor.widgets)
		      (set-fix-ink-widgets! (car cursor.widgets) '()))
		    (demo-drawing-cursor-inks drawing))))
       (%trace2 ";blinked off\n")
       (sleep-current-thread 500)
       ;; On!
       (with-glib-lock
	(lambda ()
	  (for-each (lambda (cursor.widgets)
		      (set-fix-ink-widgets! (car cursor.widgets)
					    (cdr cursor.widgets)))
		    (demo-drawing-cursor-inks drawing))))
       (%trace2 ";blinked on\n")
       (sleep-current-thread 500)
       (if (any (lambda (cursor.widgets)
		  (any (lambda (w) (not (gtk-widget-destroyed? w)))
		       (cdr cursor.widgets)))
		(demo-drawing-cursor-inks drawing))
	   (loop)
	   (%trace ";blinking ended\n"))))
   (list 'blink drawing)))


(define %trace? #f)

(define-syntax %trace
  (syntax-rules ()
    ((_ ARGS ...)
     (if %trace? (outf-error ARGS ...)))))

(define %trace2? #f)

(define-syntax %trace2
  (syntax-rules ()
    ((_ ARGS ...)
     (if %trace2? (outf-error ARGS ...)))))