#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a Gtk plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; An event viewer, a translation of Havoc Pennington's GtkEv example.
;;; package: (gtk event-viewer)

(define (make-gtk-event-viewer-demo)
  (with-glib-lock
   (lambda ()
     (let ((window (gtk-window-new 'toplevel))
	   (gtk-ev (make-gtk-event-viewer)))
       (gtk-window-set-default-size window 450 300)
       (gtk-container-add window gtk-ev)
       (gtk-window-set-title window "gtk-event-viewer")
       (gtk-container-set-border-width window 10)
       (gtk-widget-show-all window)
       gtk-ev))))

(define-class (<gtk-event-viewer> (constructor ()))
    (<scm-widget>)

  ;; GdkWindow alien, and the window geometry (allocation).
  (window define standard
	  initializer (lambda () (make-alien '|GdkWindow|)))
  (geometry define standard
	    initializer make-fix-rect)

  ;; GdkWindow alien, and the window geometry (computed from the allocation).
  (event-window define standard
		initializer (lambda () (make-alien '|GdkWindow|)))
  (event-box define standard
	     initializer make-fix-rect)

  ;; Geometry of the description area.
  (description-box define standard
		   initializer make-fix-rect)

  ;; List of lines (strings, no newlines) to be displayed in the
  ;; description area.
  (buffer define standard
	  initial-value '()))

(define-method initialize-instance ((widget <gtk-event-viewer>))
  (call-next-method widget)
  (%trace ";\t(initialize-instance <gtk-event-viewer>) "widget"\n")
  (assert-glib-locked '(initialize-instance <gtk-event-viewer>))
  (let ((alien (gobject-alien widget)))
    (C-call "gtk_widget_set_has_window" alien 1)
    (C-call "gtk_widget_set_can_focus" alien 1))
  (set-gtk-widget-size-allocate-callback! widget size-allocate-callback)
  (set-gtk-widget-realize-callback! widget realize-callback)
  (set-gtk-widget-unrealize-callback! widget unrealize-callback)
  (set-gtk-widget-draw-callback! widget draw-callback)
  (set-gtk-widget-event-callback! widget event-callback))

(define (realize-callback widget)
  (%trace2 ";realize "widget"\n")
  (assert-glib-locked 'realize-callback)
  (let ((alien (gobject-alien widget))
	(attr (malloc (C-sizeof "GdkWindowAttr") '|GdkWindowAttr|))
	(main-GdkWindow (gtk-event-viewer-window widget))
	(event-GdkWindow (gtk-event-viewer-event-window widget))
	(parent-GdkWindow (make-alien '|GdkWindow|))
	(GdkCursor (make-alien '|GdkCursor|)))

    ;; Main widget window.
    (C->= attr "GdkWindowAttr window_type" (C-enum "GDK_WINDOW_CHILD"))
    (let ((b (gtk-event-viewer-geometry widget)))
      (C->= attr "GdkWindowAttr x" (fix-rect-x b))
      (C->= attr "GdkWindowAttr y" (fix-rect-y b))
      (C->= attr "GdkWindowAttr width" (fix-rect-width b))
      (C->= attr "GdkWindowAttr height" (fix-rect-height b)))
    (C->= attr "GdkWindowAttr wclass" (C-enum "GDK_INPUT_OUTPUT"))
    (C->= attr "GdkWindowAttr event_mask"
	  (bit-ior (C-call "gtk_widget_get_events" alien)
		   (C-enum "GDK_EXPOSURE_MASK")))

    (C-call "gtk_widget_get_parent_window" parent-GdkWindow alien)
    (error-if-null parent-GdkWindow "Could not get parent:" widget)

    (C-call "gdk_window_new" main-GdkWindow parent-GdkWindow attr
	    (bit-ior (C-enum "GDK_WA_X") (C-enum "GDK_WA_Y")))
    (error-if-null main-GdkWindow "Could not create main window:" widget)
    (C-call "gtk_widget_set_window" alien main-GdkWindow)
    (C-call "gdk_window_set_user_data" main-GdkWindow alien)

    ;; Event window
    (let ((GdkDisplay (make-alien '|GdkDisplay|)))
      (C-call "gtk_widget_get_display" GdkDisplay alien)
      (C-call "gdk_cursor_new_for_display" GdkCursor
	      GdkDisplay (C-enum "GDK_CROSSHAIR")))
    (error-if-null GdkCursor "Could not create cursor:" widget)
    (C->= attr "GdkWindowAttr window_type" (C-enum "GDK_WINDOW_CHILD"))
    (let ((b (gtk-event-viewer-event-box widget)))
      (C->= attr "GdkWindowAttr x" (fix-rect-x b))
      (C->= attr "GdkWindowAttr y" (fix-rect-y b))
      (C->= attr "GdkWindowAttr width" (fix-rect-width b))
      (C->= attr "GdkWindowAttr height" (fix-rect-height b)))
    (C->= attr "GdkWindowAttr wclass" (C-enum "GDK_INPUT_OUTPUT"))
    (C->= attr "GdkWindowAttr event_mask" (C-enum "GDK_ALL_EVENTS_MASK"))
    (C->= attr "GdkWindowAttr cursor" GdkCursor)
    (C-call "gdk_window_new" event-GdkWindow main-GdkWindow attr
	    (bit-ior (C-enum "GDK_WA_X") (C-enum "GDK_WA_Y")
		     (C-enum "GDK_WA_CURSOR")))
    (free attr)
    (error-if-null event-GdkWindow "Could not create event window:" widget)
    (C-call "gdk_window_set_user_data" event-GdkWindow alien)
    (C-call "gdk_window_show" event-GdkWindow)
    (C-call "g_object_unref" GdkCursor)

    unspecific))

(define (unrealize-callback widget)
  (%trace2 ";unrealize "widget"\n")
  (assert-glib-locked 'unrealize-callback)
  ;; Destroy our event window.
  (let ((event-GdkWindow (gtk-event-viewer-event-window widget)))
    (if (not (alien-null? event-GdkWindow))
	(begin
	  (C-call "gdk_window_set_user_data" event-GdkWindow 0)
	  (C-call "gdk_window_destroy" event-GdkWindow)
	  (alien-null! event-GdkWindow))))
  ;; The main window will be destroyed by the toolkit.
  unspecific)

(define (size-allocate-callback widget GtkAllocation)
  (%trace2 ";size-allocate "widget"\n")
  (assert-glib-locked 'size-allocate-callback)
  (let ((x (C-> GtkAllocation "GtkAllocation x"))
	(y (C-> GtkAllocation "GtkAllocation y"))
	(width (C-> GtkAllocation "GtkAllocation width"))
	(height (C-> GtkAllocation "GtkAllocation height"))
	(spacing 10))
    (%trace2 ";\t"x","y" "width"x"height"\n")
    (set-fix-rect! (gtk-event-viewer-geometry widget) x y width height)
    (let ((event-width (max (- width (* 2 spacing)) 0))
	  (event-height (max (- (quotient height 5) spacing) 0)))
      (let ((event-x (quotient (- width event-width) 2))
	    (event-y (min height spacing)))
	(set-fix-rect! (gtk-event-viewer-event-box widget)
		       event-x event-y event-width event-height)
	(let* ((desc-x event-x)
	       (desc-y (+ event-y (+ event-height spacing)))
	       (desc-width event-width)
	       (desc-height (max (- height (+ event-height (* 3 spacing)))
				 0)))
	  (set-fix-rect! (gtk-event-viewer-description-box widget)
			 desc-x desc-y desc-width desc-height))

	(if (gtk-widget-realized? widget)
	    (begin
	      (C-call "gdk_window_move_resize"
		      (gtk-event-viewer-window widget) x y width height)
	      (C-call "gdk_window_move_resize"
		      (gtk-event-viewer-event-window widget)
		      event-x event-y event-width event-height)))
	unspecific))))

(define (event-callback widget GdkEvent)
  (%trace2 ";event-callback "widget" "GdkEvent"\n")
  (let ((window (C-> GdkEvent "GdkEvent any window"))
	(type (C-> GdkEvent "GdkEvent any type")))
    (%trace "; "(C-enum "GdkEventType" type)
	   " on window 0x"(alien/address-string window)".\n")

    (push-text widget (event-to-text GdkEvent))

    (if (and (= type (C-enum "GDK_KEY_PRESS"))
	     (= (C-> GdkEvent "GdkEvent key keyval") (C-enum "GDK_KEY_D")))
	;; Test debugging in a callback.
	(bkpt 'Test))

    ;; Unnecessary and deprecated.
    #;(if (= type (C-enum "GDK_MOTION_NOTIFY"))
	(C-call "gdk_window_get_pointer" #f
		(C-> GdkEvent "GdkEventMotion window") 0 0 0))
    #f ;; continue handling.
    ))

(define (draw-callback widget cr)
  (%trace ";  Draw "widget"\n")
  (set-alien/ctype! cr '|cairo_t|)
  (paint-event-window widget cr)
  (paint-window widget cr)
  #t)

(define (paint-window widget cr)
  (%trace2 ";(paint-window "widget" "cr")\n")
  (assert-glib-locked 'paint-window)
  (let ((alien (gobject-alien widget))
	(style (make-alien '|GtkStyleContext|))
	(event-box (gtk-event-viewer-event-box widget)))

    (cairo-rectangle cr
		     (-1+ (fix-rect-x event-box))
		     (-1+ (fix-rect-y event-box))
		     (+ 2 (fix-rect-width event-box))
		     (+ 2 (fix-rect-height event-box)))
    (cairo-stroke cr)

    (C-call "gtk_widget_get_style_context" style alien)

    (if (gtk-widget-has-focus? widget)
	(C-call "gtk_render_focus" style cr
		(->flonum (fix:- (fix-rect-x event-box) 5))
		(->flonum (fix:- (fix-rect-y event-box) 5))
		(->flonum (fix:+ (fix-rect-width event-box) 10))
		(->flonum (fix:+ (fix-rect-height event-box) 10))))

    (if (gtk-event-viewer-buffer widget)
	(let ((descrip-box (gtk-event-viewer-description-box widget)))
	  (let ((desc-bottom (fix-rect-max-y descrip-box))
		(space 2)
		(layout (gtk-widget-create-pango-layout widget)))
	    (cairo-rectangle cr
			     (fix-rect-x descrip-box)
			     (fix-rect-y descrip-box)
			     (fix-rect-width descrip-box)
			     (fix-rect-height descrip-box))
	    (cairo-clip cr)
	    (let loop ((y (fix-rect-y descrip-box))
		       (lines (gtk-event-viewer-buffer widget)))
	      (if (null? lines)
		  unspecific
		  (begin
		    (pango-layout-set-text layout (car lines))
		    (let ((baseline (pango-layout-get-baseline layout)))
		      (cairo-move-to cr 10. y)
		      (C-call "gtk_render_layout" style cr
			      10. (->flonum y) (gobject-alien layout))
		      (let ((new-y (fix:+ y (fix:+ baseline space))))
			(if (fix:> new-y desc-bottom)
			    (begin
			      (set-cdr! lines '())
			      unspecific)
			    (loop new-y (cdr lines))))))))
	    (gobject-unref! layout))))))

(define (paint-event-window widget cr)
  (%trace2 ";(paint-event-window "widget" "cr")\n")
  (let* ((event-window (gtk-event-viewer-event-window widget))
	 (layout (gtk-widget-create-pango-layout widget ""))
	 (title (string-append "Event Window (0x"
			       (alien/address-string event-window)")")))

    (cairo-save cr)
    (cairo-set-source-color cr (->color "white" 'paint-event-window))
    (cairo-paint cr)
    (cairo-restore cr)

    (pango-layout-set-text layout title)
    (pango-layout-get-pixel-extents
     layout
     (lambda (width height)
       (declare (ignore height))
       (cairo-move-to cr
		      ;;center
		      (quotient (- (fix-rect-width
				    (gtk-event-viewer-event-box widget))
				   width)
				2)
		      10.)))
    (cairo-show-pango-layout cr layout)
    (gobject-unref! layout)
    unspecific))

(define (push-text ev lines)
  (assert-glib-locked 'push-text)
  (set-gtk-event-viewer-buffer! ev (append lines (gtk-event-viewer-buffer ev)))
  (if (gtk-widget-drawable? ev)
      (let ((a (gobject-alien ev))
	    (r (gtk-event-viewer-description-box ev)))
	(C-call "gtk_widget_queue_draw_area"
		a (fix-rect-x r) (fix-rect-y r)
		(fix-rect-width r) (fix-rect-height r)))))


(define (event-to-text GdkEvent)
  (let ((name-line (event-name-line GdkEvent))
	(any-line (any-event-line GdkEvent))
	(detail (event-detail-line GdkEvent))
	(state (event-state-line GdkEvent)))
    (append!
     (list name-line)
     (list (string-append "  " any-line))
     (if detail (list (string-append "  " detail)) '())
     (if state (list (string-append "  " state)) '()))))

(define (event-name-line GdkEvent)
  (let ((type (C-> GdkEvent "GdkEvent any type")))
    (string-append (symbol->string (C-enum "GdkEventType" type)) "\n")))

(define (any-event-line GdkEvent)
  (assert-glib-locked 'any-event-line)
  (let ((event-time (C-call "gdk_event_get_time" GdkEvent))
	(addr (alien/address-string (C-> GdkEvent "GdkEvent any window")))
	(send (if (not (= 0 (C-> GdkEvent "GdkEvent any send_event")))
		  "True" "False")))
    (if (not (= event-time (C-enum "GDK_CURRENT_TIME")))
	(cat "Window: 0x"addr" Time: "event-time" send_event: "send"\n")
	(cat "Window: 0x"addr" send_event: "send"\n"))))

(define (cat . objects)
  (apply string-append (map (lambda (obj)
			      (if (string? obj) obj (write-to-string obj)))
			    objects)))

(define (event-state-line GdkEvent)
  (let* ((type (C-> GdkEvent "GdkEvent any type"))
	 (state	;;GdkModifierType
	  (cond ((= type (C-enum "GDK_MOTION_NOTIFY"))
		 (C-> GdkEvent "GdkEvent motion state"))
		((memv type `(,(C-enum "GDK_BUTTON_PRESS")
			      ,(C-enum "GDK_2BUTTON_PRESS")
			      ,(C-enum "GDK_3BUTTON_PRESS")
			      ,(C-enum "GDK_BUTTON_RELEASE")))
		 (C-> GdkEvent "GdkEvent button state"))
		((memv type `(,(C-enum "GDK_KEY_PRESS")
			      ,(C-enum "GDK_KEY_RELEASE")))
		 (C-> GdkEvent "GdkEvent key state"))
		(else #f)))
	 (line
	  (and state
	       (string-append
		(decorated-string-append
		 "" " | " ""
		 (append!
		  (if (bit? state (C-enum "GDK_SHIFT_MASK"))
		      (list "Shift") '())
		  (if (bit? state (C-enum "GDK_LOCK_MASK"))
		      (list "Lock") '())
		  (if (bit? state (C-enum "GDK_CONTROL_MASK"))
		      (list "Ctrl") '())
		  (if (bit? state (C-enum "GDK_MOD1_MASK"))
		      (list "Mod1") '())
		  (if (bit? state (C-enum "GDK_MOD2_MASK"))
		      (list "Mod2") '())
		  (if (bit? state (C-enum "GDK_MOD3_MASK"))
		      (list "Mod3") '())
		  (if (bit? state (C-enum "GDK_MOD4_MASK"))
		      (list "Mod4") '())
		  (if (bit? state (C-enum "GDK_MOD5_MASK"))
		      (list "Mod5") '())
		  (if (bit? state (C-enum "GDK_BUTTON1_MASK"))
		      (list "Button1") '())
		  (if (bit? state (C-enum "GDK_BUTTON2_MASK"))
		      (list "Button2") '())
		  (if (bit? state (C-enum "GDK_BUTTON3_MASK"))
		      (list "Button3") '())
		  (if (bit? state (C-enum "GDK_BUTTON4_MASK"))
		      (list "Button4") '())
		  (if (bit? state (C-enum "GDK_RELEASE_MASK"))
		      (list "Release") '())))
		"\n"))))
    (if (or (not line) (string=? line "\n")) #f line)))

(define (event-detail-line GdkEvent)
  (let ((type (C-> GdkEvent "GdkEvent any type")))
    (cond ((= type (C-enum "GDK_MOTION_NOTIFY"))
	   (let ((x (C-> GdkEvent "GdkEvent motion x"))
		 (y (C-> GdkEvent "GdkEvent motion y")))
	     (cat "x: "x" y: "y"\n")))
	  ((memq type `(,(C-enum "GDK_BUTTON_PRESS")
			,(C-enum "GDK_2BUTTON_PRESS")
			,(C-enum "GDK_3BUTTON_PRESS")
			,(C-enum "GDK_BUTTON_RELEASE")))
	   (cat "Button: "(C-> GdkEvent "GdkEvent button button")"\n"))
	  ((memq type `(,(C-enum "GDK_KEY_PRESS")
			,(C-enum "GDK_KEY_RELEASE")))
	   (let ((keyval (C-enum "enum GdkKeysyms"
				 (C-> GdkEvent "GdkEvent key keyval")))
		 (text (let ((alien (make-alien '|gchar|)))
			 (C-> GdkEvent "GdkEvent key string" alien)
			 (c-peek-cstring alien))))
	     (cat "Keyval: "keyval" Text: "text"\n")))
	  (else
	   #f))))

(define %trace? #f)

(define-syntax %trace
  (syntax-rules ()
    ((_ ARGS ...)
     (if %trace? (outf-error ARGS ...)))))

(define %trace2? #f)

(define-syntax %trace2
  (syntax-rules ()
    ((_ ARGS ...)
     (if %trace2? (outf-error ARGS ...)))))