#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a Gtk plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; A <gtk-widget> representing a ScmWidget.
;;; package: (gtk widget)

(define-class <scm-widget> (<gtk-widget>))

(define-guarantee scm-widget "a <scm-widget>")

(define-method initialize-instance ((new <scm-widget>))
  (call-next-method new)
  (assert-glib-locked '(initialize-instance <scm-widget>))
  (let ((a (gobject-alien new)))
    (C-call "scm_widget_new" a)
    (if (alien-null? a)
	(error "Could not create: a Scheme widget.")
	(C-call "g_object_ref_sink" a a))
    (set-gtk-widget-destroy-callback! new)))

(define (set-scm-widget-set-scroll-adjustments-callback! widget callback)
  (guarantee-scm-widget widget 'set-scm-widget-set-scroll-adjustments-callback!)
  (guarantee-procedure-of-arity callback 3 'set-scm-widget-set-scroll-adjustments-callback!)
  (g-signal-connect widget (C-callback "set_scroll_adjustments") callback))

(define (set-scm-widget-minimum-size! widget width height)
  (guarantee-scm-widget widget 'set-scm-widget-minimum-size!)
  (guarantee non-negative-fixnum? width 'set-scm-widget-minimum-size!)
  (guarantee non-negative-fixnum? height 'set-scm-widget-minimum-size!)
  (let ((a (gobject-alien widget)))
    (C->= a "ScmWidget minimum_width" width)
    (C->= a "ScmWidget minimum_height" height)))

(define (set-scm-widget-natural-size! widget width height)
  (guarantee-scm-widget widget 'set-scm-widget-natural-size!)
  (guarantee non-negative-fixnum? width 'set-scm-widget-natural-size!)
  (guarantee non-negative-fixnum? height 'set-scm-widget-natural-size!)
  (let ((a (gobject-alien widget)))
    (C->= a "ScmWidget natural_width" width)
    (C->= a "ScmWidget natural_height" height)))